import requests
import os

from .error import APIError
from .session import Environment
from pyrasgo.schemas.user import UserRegistration, UserLogin

class Register:

    def login(self, payload: UserLogin):
        url = self._url(resource=f"/pyrasgo-login", api_version=1)
        response = requests.post(url, json=payload.__dict__)
        status_code = response.status_code
        if status_code == 200:
            return response.json()
        if status_code == 401 or status_code == 403:
            raise APIError("Username or password are incorrect. Please check your credentials "
                    "or use pyrasgo.register() to create a new account.")
        elif status_code == 400:
            raise APIError("Credentials Expired. Contact Rasgo Support.")

    def register(self, payload: UserRegistration):
        url = self._url(resource=f"/pyrasgo-register", api_version=1)
        response = requests.post(url, json=payload.__dict__)
        status_code = response.status_code
        if status_code == 200:
            return response.json()
        elif status_code == 401:
            raise APIError("Unable to register user. Password must be at least 6 characters.")
        else:
            raise APIError("Unable to register user. Please register with valid credentials "
                    "or use pyrasgo.login() if already registered.")

    def _url(self, resource, api_version=None):
        env = Environment.from_environment()
        if '/' == resource[0]:
            resource = resource[1:]
        protocol = 'http' if env.value == 'localhost' else 'https'
        return f"{protocol}://{env.value}/{'' if api_version is None else f'v{api_version}/'}{resource}"



import pandas as pd
from typing import List, Optional

from .error import APIError
from pyrasgo import schemas
from pyrasgo import primitives


class Get():

    def __init__(self):
        from .connection import Connection
        from pyrasgo.config import get_session_api_key
        from pyrasgo.storage import DataWarehouse, SnowflakeDataWarehouse

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)
        self.data_warehouse: SnowflakeDataWarehouse = DataWarehouse.connect()

    def collection(self, id: int) -> primitives.Collection:
        """
        Returns a Rasgo Collection (set of joined Features) matching the specified id
        """
        try:
            return primitives.Collection(api_object=self.api._get(f"/models/{id}", api_version=1).json())
        except:
            raise APIError(f"Collection {id} does not exist or this API key does not have access.")

    def collection_attributes(self, id: int) -> schemas.CollectionAttributes:
        """
        Returns a dict of attributes for a collection
        """
        try:
            response = self.api._get(f"/models/{id}/attributes", api_version=1).json()
            dict_out = {}
            for kv in response:
                dict_out[kv['key']] = kv['value']
            return schemas.CollectionAttributes(collectionId=id, attributes=dict_out)
        except:
            raise APIError(f"Collection {id} does not exist or this API key does not have access.")

    def collections(self, include_shared: bool=False) -> List[primitives.Collection]:
        """
        Returns all Rasgo Collections (set of joined Features) that I have author access to. Add an include_shared
        parameter to return all Rasgo Collections that I have any access to (author or shared access)
        :param include_shared: Boolean value indicating if the return should include all accessible collections
        """
        try:
            return [primitives.Collection(api_object=entry) for entry in self.api._get(f"/models", {"include_shared": include_shared}, api_version=1).json()]
        except:
            raise APIError("Collections do not exist or this API key does not have access.")

    def collections_by_attribute(self, key: str, value: str = None) -> List[primitives.Collection]:
        """
        Returns a list of Rasgo Collections that match an attribute
        """
        try:
            params = {"key": key}
            if value:
                params.update({"value": value})
            response = self.api._get(f"/models/attributes/models", params=params, api_version=1).json()
            return [primitives.Collection(api_object=r) for r in response]
        except:
            raise APIError(f"Key {key}: {value or 'Any'} does not exist or this API key does not have access.")

    def data_sources(self, with_features_only: bool = False) -> List[primitives.DataSource]:
        """
        Returns all DataSources available in your organization or Rasgo Community
        """
        try:
            response = self.api._get("/data-source", {"with_features_only": with_features_only}, api_version=1).json()
            return [primitives.DataSource(api_object=entry) for entry in response]
        except:
            raise APIError("Data Sources do not exist or this API key does not have access.")

    def data_source(self, id: int) -> primitives.DataSource:
        """
        Returns the DataSource with the specified id
        """
        try:
            response = self.api._get(f"/data-source/{id}", api_version=1).json()
            return primitives.DataSource(api_object=response)
        except:
            raise APIError(f"Data Source {id} does not exist or this API key does not have access.")

    def data_source_stats(self, id: int):
        """
        Returns the stats profile of the specificed data source
        """
        try:
            return self.api._get(f"/data-source/profile/{id}", api_version=1).json()
        except:
            raise APIError(f"Stats do not exist for DataSource {id}")

    def dataframes(self) -> List[schemas.Dataframe]:
        """
        Returns all Dataframes available in your organization or Rasgo Community
        """
        try:
            response = self.api._get("/dataframes", api_version=1).json()
            return [schemas.Dataframe(**entry) for entry in response]
        except:
            raise APIError("Dataframes do not exist or this API key does not have access.")

    def dataframe(self, unique_id: str) -> schemas.Dataframe:
        """
        Returns the Dataframe with the specified id
        """
        try:
            response = self.api._get(f"/dataframes/{unique_id}", api_version=1).json()
            return schemas.Dataframe(**response)
        except:
            raise APIError(f"Dataframe {unique_id} does not exist or this API key does not have access.")

    def feature(self, id: int) -> primitives.Feature:
        """
        Returns the Feature with the specified id
        """
        try:
            return primitives.Feature(api_object=self.api._get(f"/features/{id}", api_version=1).json())
        except:
            raise APIError(f"Feature {id} does not exist or this API key does not have access.")

    def feature_attributes(self, id: int) -> schemas.FeatureAttributes:
        """
        Returns a dict of attributes for a feature
        """
        try:
            response = self.api._get(f"/features/{id}/attributes", api_version=1).json()
            dict_out = {}
            for kv in response:
                dict_out[kv['key']] = kv['value']
            return schemas.FeatureAttributes(featureId=id, attributes=dict_out)
        except:
            raise APIError(f"Feature {id} does not exist or this API key does not have access.")

    def feature_attributes_log(self, id: int) -> tuple:
        """
        Returns a list of all attributes values logged to a feature over time
        """
        try:
            response = self.api._get(f"/features/{id}/attributes/log", api_version=1).json()
            lst_out = []
            for kv in response:
                dict_item={}
                dict_item[kv['key']] = kv.get('value', None)
                dict_item['updatedBy'] = kv.get('recordAuthorId', None)
                dict_item['updated'] = kv.get('recordTimestamp', None)
                lst_out.append(dict_item)
            return schemas.FeatureAttributesLog(featureId=id, attributes=lst_out)
        except:
            raise APIError(f"Feature {id} does not exist or this API key does not have access.")


    def features_yml(self, data_source_id: int) -> schemas.FeaturesYML:
        """
        Returns the Features with the specified id
        """
        try:
            response = self.data_source(data_source_id)
            return schemas.FeaturesYML(
                name = response.name,
                sourceTable = response.dataTable.fqtn,
                dimensions = [{"columnName": d.columnName, "dataType": d.dataType, "granularity": d.granularity.name} for d in response.dimensions],
                features = response.features,
                sourceCode = response.sourceCode,
                sourceType = response.sourceType
            )
        except:
            raise APIError(f"Feature {id} does not exist or this API key does not have access.")

    def feature_stats(self, id: int) -> Optional[schemas.FeatureStats]:
        """
        Returns the stats profile for the specified Feature
        """
        try:
            stats_json = self.api._get(f"/features/{id}/stats", api_version=1).json()
            return schemas.FeatureStats(**stats_json["featureStats"])
        except:
            raise APIError(f"Stats do not exist yet for feature {id}.")

    def features(self) -> primitives.FeatureList:
        """
        Returns a list of Features available in your organization or Rasgo Community
        """
        try:
            return primitives.FeatureList(api_object=self.api._get("/features", api_version=1).json())
        except:
            raise APIError("Features do not exist or this API key does not have access.")

    def features_by_attribute(self, key: str, value: str = None) -> primitives.FeatureList:
        """
        Returns a list of features that match an attribute
        """
        try:
            params = {"key": key}
            if value:
                params.update({"value": value})
            return primitives.FeatureList(api_object=self.api._get(f"/features/attributes/features", params=params, api_version=1).json())
        except:
            raise APIError(f"Key {key}: {value or 'Any'} does not exist or this API key does not have access.")

    def shared_collections(self) -> List[primitives.Collection]:
        """
        Returns all Rasgo Collections (set of joined Features) shared in my organization or in Rasgo community
        """
        try:
            return [primitives.Collection(api_object=entry) for entry in self.api._get(f"/models/shared", api_version=1).json()]
        except:
            raise APIError("Shared Collections do not exist or this API key does not have access.")

    def source_columns(self, table: Optional[str] = None, database: Optional[str] = None, schema: Optional[str] = None, data_type: Optional[str] = None) -> pd.DataFrame:
        """
        Returns a DataFrame of columns in Snowflake tables and views that are queryable as feature sources
        """
        return self.data_warehouse.get_source_columns(table=table, database=database, schema=schema, data_type=data_type)

    def source_tables(self, database: Optional[str] = None, schema: Optional[str] = None) -> pd.DataFrame:
        """
        Return a DataFrame of Snowflake tables and views that are queryable as feature sources
        """
        return self.data_warehouse.get_source_tables(database=database, schema=schema)

    def transform(self, transform_id: int) -> schemas.Transform:
        """Returns an individual transform"""
        try:
            response = self.api._get(f"/transform/{transform_id}", api_version=1).json()
            return schemas.Transform(**response)
        except:
            raise APIError(f"Transform with id '{transform_id}' does not exist "
                           f"or this API key does not have access.")

    def transforms(self) -> List[schemas.Transform]:
        """Returns a list of available transforms
        """
        response = self.api._get(f"/transform", api_version=1).json()
        return [schemas.Transform(**r) for r in response]

    def user(self):
        response = self.api._get("/users/me", api_version=1).json()
        return schemas.User(**response)

    def dataset(
            self,
            dataset_id: Optional[int] = None,
            fqtn: Optional[str] = None,
            resource_key: Optional[str] = None,
            version: Optional[int] = None
    ) -> primitives.Dataset:
        """
        Return a Rasgo dataset primitive by Id or FQTN
        """
        no_input_supplied = dataset_id is None and not fqtn and not resource_key
        too_many_inputs_supplied = (bool(dataset_id) + bool(fqtn) + bool(resource_key or version)) > 1
        if no_input_supplied or too_many_inputs_supplied:
            raise ValueError("Valid Dataset retrieval options are:\n"
                             "1. id\n2. fqtn\n3. resource_key\n4. resource_key and version")

        try:
            if dataset_id:
                response = self.api._get(f"/datasets/{dataset_id}", api_version=2).json()
            elif fqtn:
                response = self.api._get(f"/datasets/match/{fqtn}", api_version=2).json()
            elif resource_key:
                response = self.api._get(
                    f"/datasets/rk/{resource_key}{f'/{version}' if version else ''}",
                    api_version=2
                ).json()
            dataset_schema = schemas.Dataset(**response)

            operation_set_schema = None
            if dataset_schema.dw_operation_set_id:
                response = self.api._get(
                    f"/operation-sets/{dataset_schema.dw_operation_set_id}",
                    api_version=2
                ).json()
                operation_set_schema = schemas.OperationSet(**response)
            return primitives.Dataset(
                api_dataset=dataset_schema,
                api_operation_set=operation_set_schema
            )
        except:
            if dataset_id:
                raise APIError(f"Dataset with id '{dataset_id!r}' does not exist "
                               f"or this API key does not have access.")
            elif fqtn:
                raise APIError(f"Dataset with fqtn '{fqtn}' does not exist "
                               f"or this API key does not have access.")

    def dataset_yaml(
            self,
            dataset_id: Optional[int] = None,
    ) -> str:
        try:
            return self.api._get(f"/datasets/{dataset_id}/yaml", api_version=2).text
        except:
            raise APIError(f"Dataset with id '{dataset_id!r}' does not exist or this API key does not have access.")

    def datasets(self) -> List[primitives.Dataset]:
        """
        Return all datasets in Rasgo attached to your organization
        """
        # Get transforms  so we can cache them for use in transforming datasets
        transforms = self.transforms()
        response = self.api._get("/datasets", api_version=2).json()
        datasets = []
        for r in response:
            dataset_schema = schemas.Dataset(**r)
            dataset = primitives.Dataset(
                api_dataset=dataset_schema,
                transforms=transforms
            )
            datasets.append(dataset)
        return datasets


    # ----------------------------------
    #  Internal/Private Get Calls
    # ----------------------------------


    def _operation_set(self, operation_set_id: int) -> schemas.OperationSet:
        """
        Return a Rasgo operation set by id
        """
        response = self.api._get(
            f"/operation-sets/{operation_set_id}",
            api_version=2
        ).json()
        return schemas.OperationSet(**response)

    def _operation_set_async_status(self, task_id: int) -> schemas.OperationSetAsyncTask:
        """
        Returns the status of an Operation Set creation task by id
        """
        response = self.api._get(
            f"/operation-sets/async/{task_id}",
            api_version=2
        ).json()
        return schemas.OperationSetAsyncTask(**response)

    def _operation_set_offline_async_status(self, task_id: int) -> schemas.OperationSetOfflineAsyncTask:
        """
        Returns the status of an Operation Set creation task by id
        """
        response = self.api._get(
            f"/operation-sets/offline/async/{task_id}",
            api_version=2
        ).json()
        return schemas.OperationSetOfflineAsyncTask(**response)



from typing import List

class APIError(Exception):
    pass

class APIWarning(Warning):
    pass

class ParameterValueError(Exception):
    def __init__(self, param: str, values: List[str]):
        self.param = param
        self.values = values

    def __str__(self):
        return f"Please pass a valid value for {self.param}. Values are {self.values}"



"""
API Contracts for Data Warehouse Operations
"""
from typing import Dict, List, Optional, Any

from pydantic import BaseModel, Field

from pyrasgo.schemas import dw_table as dw_table_schemas


class OperationCreate(BaseModel):
    """
    Contract to create an Operation on set
    """
    operation_name: str = Field(alias='operationName')
    operation_args: Dict = Field(alias='operationArgs')
    transform_id: int = Field(alias='transformId')
    table_dependency_names: Optional[List[str]] = Field(alias="tableDependencyNames") # DW Table names
    table_dependency_ids: Optional[List[int]] = Field(alias="tableDependencyIds") # DW Table ids
    table_name: Optional[str] = Field(alias="tableName")

    class Config:
        allow_population_by_field_name = True


class OperationUpdate(BaseModel):
    """
    Contract to update an Operation on set
    """
    operation_name: Optional[str] = Field(alias='operationName')
    operation_args: Optional[Dict] = Field(alias='operationArgs')
    transform_id: Optional[int] = Field(alias='transformId')
    table_dependency_names: Optional[List[str]] = Field(alias="tableDependencyNames") # DW Table names
    table_dependency_ids: Optional[List[int]] = Field(alias="tableDependencyIds") # DW Table ids
    dw_operation_set_id: Optional[int] = Field(alias='dwOperationSetId')

    class Config:
        allow_population_by_field_name = True


class Operation(BaseModel):
    """
    Contract representing a Operation
    """
    id: int
    operation_name: str = Field(alias='operationName')
    operation_args: Any = Field(alias='operationArgs')
    operation_sql: str = Field(alias='operationSQL')
    transform_id: int = Field(alias='transformId')
    dw_table_id: int = Field(alias='dwTableId')
    dw_operation_set_id: int = Field(alias='dwOperationSetId')
    dw_table: Optional[dw_table_schemas.DataTableWithColumns] = Field(alias='dataTable')
    table_dependencies: Optional[List[int]] = Field(alias="tableDependencies") # DW Table ids

    class Config:
        allow_population_by_field_name = True


class OperationAdjacency(BaseModel):
    """
    Contract representing a Operation Adjacency
    """
    direction: str
    udt_operation_id: int = Field(alias='operationId')
    dw_table_id: int = Field(alias='dwTableId')




class Admin():

    def __init__(self):
        from .connection import Connection
        from pyrasgo.config import get_session_api_key

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)

    def fix_permission_errors(self, user_id: int):
        response = self.api._patch(f"/admin/users/{user_id}/fix-permission-errors", api_version=1).json()
        return response


from typing import List, Optional

from .error import APIError
from pyrasgo import schemas as api
from pyrasgo.primitives import Collection, Feature, FeatureList, DataSource
from pyrasgo.schemas.enums import Granularity, ModelType


class Match():

    def __init__(self):
        from . import Get
        from .connection import Connection
        from pyrasgo.config import get_session_api_key

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)
        self.get = Get()

    def data_source(self, fqtn: str) -> DataSource:
        """
        Returns the first Data Source that matches the specified name
        """
        try:
            response = self.api._get(f"/data-source", {"table": fqtn}, api_version=1).json()
            return DataSource(api_object=response[0])
        except:
            return None

    def dataframe(self, name: str = None, unique_id: str = None) -> api.Dataframe:
        """
        Returns the first Dataframe that matches the specified name or uid
        """
        try:
            if name:
                response = self.api._get(f"/dataframes", {"name": name}, api_version=1).json()
                return api.Dataframe(**response[0])
            elif unique_id:
                response = self.api._get(f"/dataframes/{unique_id}", api_version=1).json()
                return api.Dataframe(**response)
        except:
            return None
        return None

    def feature(self, column_name: str, data_source_id: int) -> Optional[Feature]:
        """
        Returns the first Feature matching the specified name in a DataSource
        """
        try:
            features = self.get.data_source(data_source_id).features
            for f in features:
                if column_name.upper() == f.columnName.upper():
                    return Feature(api_object=f.dict())
            return None
        except:
            return None


import logging
import os
from collections import OrderedDict
from typing import TYPE_CHECKING, Dict, List, Optional, Union

import pandas as pd
import yaml
from pyrasgo.api.connection import Connection
from pyrasgo.schemas import DataType, FeaturesYML
from pyrasgo.schemas import data_source as schema
from pyrasgo.schemas.transform import Transform
from pyrasgo.utils import naming

# Import TransformChain for type checking only
# to avoid circular import
if TYPE_CHECKING:
    from pyrasgo.utils.transforms import TransformChain


class DataSource(Connection):
    """
    Stores a Rasgo DataSource
    """
    def __init__(self, api_object, **kwargs):
        super().__init__(**kwargs)
        self._fields = schema.DataSource(**api_object)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"Source(id={self.id}, name={self.name}, sourceType={self.sourceType}, table={self.fqtn})"

    def __getattr__(self, item):
        try:
            return self._fields.__getattribute__(item)
        except KeyError:
            self.refresh()
        try:
            return self._fields.__getattribute__(item)
        except KeyError:
            raise AttributeError(f"No attribute named {item}")

# ----------
# Properties
# ----------

    @property
    def fqtn(self) -> str:
        return naming.make_fqtn(
            table=self.table,
            database=self.tableDatabase,
            schema=self.tableSchema
        )

# --------
# Methods
# --------

    def display_source_code(self):
        """
        Convenience function to display the sourceCode property
        """
        return self._fields.sourceCode

    def read_into_df(self,
                     filters: Optional[Dict[str, str]] = None,
                     limit: Optional[int] = None) -> pd.DataFrame:
        """
        Pull Source data from DataWarehouse into a pandas Dataframe
        """
        from pyrasgo.api.read import Read
        return Read().source_data(id=self.id, filters=filters, limit=limit)

    def rebuild_from_source_code(self):
        """
        Rebuild the Source using the source code
        """
        raise NotImplementedError()

    def refresh(self):
        """
        Updates the Soure's attributes from the API
        """
        self._fields = schema.DataSource(**self._get(f"/data-source/{self.id}", api_version=1).json())

    def rename(self, new_name: str):
        """
        Updates a DataSource's display name
        """
        print(f"Renaming DataSource {self.id} from {self.name} to {new_name}")
        source = schema.DataSourceUpdate(id=self.id, name=new_name)
        self._fields = schema.DataSource(**self._patch(f"/data-source/{self.id}",
                                                    api_version=1, _json=source.dict(exclude_unset=True, exclude_none=True)).json())

    def to_dict(self) -> dict:
        for f in self._fields.features:
            attr_dict = {}
            for a in f.attributes:
                attr_dict.update({a.key: a.value})
            f.attributes = attr_dict
        return FeaturesYML(
                name=self._fields.name,
                sourceTable=self._fields.table,
                dimensions=[{"columnName": d.columnName, "dataType": d.dataType, "granularity": d.granularity.name} for d in self._fields.dimensions],
                features=self._fields.features,
                sourceCode=self._fields.sourceCode,
                sourceType=self._fields.sourceType
        ).dict(exclude_unset=False, by_alias=True)

    def to_yml(self, file_name: str,
                     directory: str = None,
                     overwrite: bool = True
    ) -> str:
        if directory is None:
            directory = os.getcwd()

        if directory[-1] == "/":
            directory = directory[:-1]

        if file_name.split(".")[-1] not in ['yaml', 'yml']:
            file_name += ".yaml"

        if os.path.exists(f"{directory}/{file_name}") and overwrite:
            logging.warning(f"Overwriting existing file {file_name} in directory: {directory}")

        safe_dumper = yaml.SafeDumper
        safe_dumper.add_representer(DataType, lambda self, data: self.represent_str(str(data.value)))
        safe_dumper.add_representer(OrderedDict, lambda self, data: self.represent_mapping('tag:yaml.org,2002:map', data.items()))
        safe_dumper.ignore_aliases = lambda self, data: True

        with open(f"{directory}/{file_name}", "w") as _yaml:
            yaml.dump(data=OrderedDict(self.to_dict()), Dumper=safe_dumper, stream=_yaml)

    def _make_table_metadata(self) -> Dict:
        return {
            "database": self.dataTable.databaseName,
            "schema": self.dataTable.schemaName,
            "table": self.dataTable.tableName,
        }

    def transform(self,
                  *,
                  transform_id: Optional[int] = None,
                  transform_name: Optional[str] = None,
                  transform: Optional[Transform] = None,
                  source_code: Optional[str] = None,
                  arguments: Optional[Dict[str, str]] = None,
                  **kwargs: Union[str, int, List, Dict]) -> 'TransformChain':
        """
        Add a transform step on this source, and return a new TransformerChain obj

        Returned obj can chain more transformers together by calling .transform() again if needed.

        Preview the transformed dataset as a df or SQL as a df with .preview() and .preview_sql()

        Create a source from the chained transforms by then calling .to_source(). You can ONLY do this
        when supplying a transform id, name, or obj and not source_code

        Args:
            transform_id: Transform id to add as step to chain. Can input Transform name or Transform schema instead
            transform_name: Transform name to add as step to chain. Can add Transform id or Transform schema instead
            transform: Transform schema to add as step to chain. Can be Transform id or name instead
            source_code: If not entering a transform id, name, or obj, test source_code with the transform
            arguments: Arguments to apply to this transform

        Returns:
            A TransformChain() obj with a newly added transform step
        """
        from pyrasgo.utils import transforms

        # Init and execute a transform chain with only one transform step
        # Error handling and validation occurs within TransformChain obj
        transform_chain = transforms.TransformChain(self)
        return transform_chain.transform(
            transform_id=transform_id,
            transform_name=transform_name,
            transform=transform,
            source_code=source_code,
            arguments=arguments,
            **kwargs
        )



import logging
import pandas as pd
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
import webbrowser

from pyrasgo import config

from pyrasgo.api import (
    Admin, Create, Delete, Get, Match, Publish, Read, Save, Update,
)
from pyrasgo.primitives import Collection, DataSource, Feature, FeatureList

from pyrasgo import schemas as api
from pyrasgo.schemas.enums import Granularity, ModelType

from pyrasgo.storage  import Evaluate, Prune, Transform
from pyrasgo.storage.dataframe.utils import generate_unique_id

from pyrasgo.utils import ingestion
from pyrasgo.utils.versioning import deprecated_with_replacement, deprecated_without_replacement


class Rasgo():
    """
    Base connection object to handle interactions with the Rasgo API.
    """
    def __init__(self):
        self._experiment_id = None
        self.get = Get()
        self.match = Match()
        self.create = Create()
        self.update = Update()
        self.publish = Publish()
        self.read = Read()
        self.delete = Delete()
        self.save = Save()
        self.transform = Transform()
        self.evaluate = Evaluate()
        self.prune = Prune()
        self.admin = Admin()

    def open_docs(self):
        webbrowser.open("https://docs.rasgoml.com/rasgo-docs/pyrasgo/pyrasgo-getting-started")

    def pronounce_rasgo(self):
        webbrowser.open("https://www.spanishdict.com/pronunciation/rasgo?langFrom=es")

    def _pyrasgo_version(self):
        from pyrasgo.version import __version__
        return __version__

    def _pyrasgo_api_key(self):
        from pyrasgo.config import get_session_api_key
        return get_session_api_key()

    def activate_experiment(
        self,
        experiment_name: str
    ):
        """
        Activate an experiment with a unique name and id. Experiments are used to
        track dataframe activity across Rasgo methods

        Note: Experiment names must be unique
              Using an existing name will re-activate an old experiment
        """
        df = self.match.dataframe(name=experiment_name)
        adjective = 'existing'
        if not df:
            uid = generate_unique_id()
            df = self.save.dataframe(name=experiment_name, unique_id=uid)
            adjective = 'new'
        self._experiment_id = df.uniqueId
        self.evaluate = Evaluate(experiment_id=self._experiment_id)
        self.prune = Prune(experiment_id=self._experiment_id)
        print(f"Activated {adjective} experiment with name {experiment_name} for dataframe: {df.uniqueId}")

    def end_experiment(self):
        """
        If an experiment is active, deactivate it and stop tracking dataframe activity
        """
        self._experiment_id = None
        self.evaluate = Evaluate()
        self.prune = Prune()
        print(f"Experiment ended")

    def is_experiment_active(self):
        """
        Check if an experiment is active
        """
        print(f"Experiment active for dataframe: {self._experiment_id}" if self._experiment_id else "No experiment active")

# ---------
# Get Calls
# ---------
    @deprecated_with_replacement('v0.3', 'get.collection')
    def get_collection(self, collection_id: int) -> Collection:
        """
        Returns a Rasgo Collection (set of joined Features) matching the specified id
        """
        return self.get.collection(collection_id)

    @deprecated_with_replacement('v0.3', 'get.collection_attributes')
    def get_collection_attributes(self, collection_id: int) -> api.CollectionAttributes:
        """
        Returns a dict of attributes for a collection
        """
        return self.get.collection_attributes(collection_id)

    @deprecated_with_replacement('v0.3', 'get.collections')
    def get_collections(self, include_shared: bool=False) -> List[Collection]:
        """
        Returns all Rasgo Collections (set of joined Features) that I have author access to. Add an include_shared
        parameter to return all Rasgo Collections that I have any access to (author or shared access)
        :param include_shared: Boolean value indicating if the return should include all accessible collections
        """
        return self.get.collections(include_shared=include_shared)

    @deprecated_with_replacement('v0.3', 'get.collections_by_attribute')
    def get_collections_by_attribute(self, key: str, value: str = None) -> List[Collection]:
        """
        Returns a list of Rasgo Collections that match an attribute
        """
        return self.get.collections_by_attribute(key, value)

    @deprecated_with_replacement('v0.3', 'get.data_sources')
    def get_data_sources(self) -> List[DataSource]:
        """
        Returns all DataSources available in your organization or Rasgo Community
        """
        return self.get.data_sources()

    @deprecated_with_replacement('v0.3', 'get.data_source')
    def get_data_source(self, data_source_id: int) -> DataSource:
        """
        Returns the DataSource with the specified id
        """
        return self.get.data_source(data_source_id)

    @deprecated_with_replacement('v0.3', 'get.data_source_stats')
    def get_data_source_stats(self, data_source_id: int):
        """
        Returns the stats profile of the specificed data source
        """
        return self.get.data_source_stats(data_source_id)

    @deprecated_with_replacement('v0.3', 'get.feature')
    def get_feature(self, feature_id: int) -> Feature:
        """
        Returns the Feature with the specified id
        """
        return self.get.feature(feature_id)

    @deprecated_with_replacement('v0.3', 'get.feature_attributes')
    def get_feature_attributes(self, feature_id: int) -> api.FeatureAttributes:
        """
        Returns a dict of attributes for a feature
        """
        return self.get.feature_attributes(feature_id)

    @deprecated_with_replacement('v0.3', 'get.feature_attributes_log')
    def get_feature_attributes_log(self, feature_id: int) -> tuple:
        """
        Returns a list of all attributes values logged to a feature over time
        """
        return self.get.feature_attributes_log(feature_id)

    @deprecated_with_replacement('v0.3', 'get.feature_stats')
    def get_feature_stats(self, feature_id: int) -> Optional[api.FeatureStats]:
        """
        Returns the stats profile for the specified Feature
        """
        return self.get.feature_stats(feature_id)

    @deprecated_with_replacement('v0.3', 'get.features')
    def get_features(self) -> FeatureList:
        """
        Returns a list of Features available in your organization or Rasgo Community
        """
        return self.get.features()

    @deprecated_with_replacement('v0.3', 'get.features_by_attribute')
    def get_features_by_attribute(self, key: str, value: str = None) -> List[Feature]:
        """
        Returns a list of features that match an attribute
        """
        return self.get.features_by_attribute(key, value)

    @deprecated_with_replacement('v0.3', 'get.shared_collections')
    def get_shared_collections(self) -> List[Collection]:
        """
        Returns all Rasgo Collections (set of joined Features) shared in my organization or in Rasgo community
        """
        return self.get.shared_collections()

    @deprecated_with_replacement('v0.3', 'get.source_columns')
    def get_source_columns(self, table: Optional[str] = None, database: Optional[str] = None, schema: Optional[str] = None, data_type: Optional[str] = None) -> pd.DataFrame:
        """
        Returns a DataFrame of columns in Snowflake tables and views that are queryable as feature sources
        """
        return self.get.source_columns(table, database, schema, data_type)

    @deprecated_with_replacement('v0.3', 'get.source_tables')
    def get_source_tables(self, database: Optional[str] = None, schema: Optional[str] = None) -> pd.DataFrame:
        """
        Return a DataFrame of Snowflake tables and views that are queryable as feature sources
        """
        return self.get.source_tables(database, schema)


# -------------------
# Post / Create Calls
# -------------------
    @deprecated_without_replacement('v0.3')
    def create_collection(self, name: str,
                          type: Union[str, ModelType],
                          granularity: Union[str, Granularity],
                          description: Optional[str] = None,
                          is_shared: Optional[bool] = False) -> Collection:
        return self.create.collection(name, type, granularity, description, is_shared)

    @deprecated_without_replacement('v0.3')
    def put_collection_attributes(self, collection_id: int, attributes: List[dict]):
        """
        Create or update attributes on a Rasgo Collection

        param attributes: dict [{"key": "value"}, {"key": "value"}]
        """
        return self.update.collection_attributes(collection_id, attributes)

    @deprecated_without_replacement('v0.3')
    def put_feature_attributes(self, feature_id: int, attributes: List[dict]):
        """
        Create or update attributes on a feature

        param attributes: dict [{"key": "value"}, {"key": "value"}]
        """
        return self.update.feature_attributes(feature_id, attributes)


# ------------------
# Workflow Functions
# ------------------
    @deprecated_with_replacement('v0.3', 'data_source.to_dict()')
    def prepare_features_dict(self, data_source_id: int = None):
        """
        Returns a dict of metadata values for the specificed Rasgo Features

        params
        ------
        data_source_id: int: ID of a Rasgo DataSource

        Alternate Usage
        ---------------
        Pass in no params to return a blank dict template
        """
        if data_source_id:
            ds = self.get.data_source(data_source_id)
            return ingestion.save_features_to_dict(ds)
        return ingestion.RASGO_DICT

    @deprecated_with_replacement('v0.3', 'data_source.to_yml()')
    def prepare_features_yml(self, data_souce_id: int, file_name: str, directory: str = None):
        """
        Saves a yml file of metadata values for the specificed Rasgo Features to a directory location

        params
        ------
        data_source_id: int: ID of a Rasgo DataSource
        file_name: str: name for the output file
        directory: str: full dir location for the output file (minus file_name)
        """
        fs = self.get.features_yml(data_souce_id)
        return ingestion.save_features_to_yaml(fs, file_name=file_name, directory=directory)

    @deprecated_with_replacement('v0.3', 'publish.features')
    def publish_features(self, features_dict: dict) -> DataSource:
        return self.publish.features(features_dict)

    @deprecated_without_replacement('v0.3')
    def publish_features_from_df(self,
                                 df: pd.DataFrame,
                                 dimensions: List[str],
                                 features: List[str],
                                 granularity: List[str] = None,
                                 tags: List[str] = None,
                                 sandbox: bool = True):
        granularity = granularity if granularity else []
        tags = tags if tags else []
        raise NotImplementedError("`publish_features_from_df()` is a deprecated method. "
                                  "Please use `publish.features()` or `publish.features_from_source()` instead")

    @deprecated_with_replacement('v0.3', 'publish.features_from_source')
    def publish_features_from_source(self,
                                     data_source_id: int,
                                     features: List[str],
                                     dimensions: List[str],
                                     granularity: List[str] = None,
                                     tags: List[str] = None,
                                     feature_set_name: str = None,
                                     sandbox: bool = True,
                                     if_exists: str = 'fail') -> DataSource:
        """
        Publishes Features from an existing DataSource table

        params:
            data_source_id: ID to a Rasgo DataSource
            features: List of column names that will be features
            dimensions: List of column names that will be dimensions
            granularity: List of grains that describe the form of the data. Common values are: 'day', 'week', 'month', 'second'
            tags: List of tags to be added to all features
            feature_set_name: deprecated parameter - any values passed in are ignored

            if_exists:  fail - returns an error message if features already exists against this table
                        return - returns the features without operating on them
                        edit - edits the existing features
                        new - creates new features

        return:
            Rasgo DataSource
        """
        granularity = granularity if granularity else []
        tags = tags if tags else []

        return self.publish.features_from_source(data_source_id, features, dimensions, granularity, tags, feature_set_name, sandbox, if_exists)

    @deprecated_with_replacement('v0.3', 'publish.features_from_yml')
    def publish_features_from_yml(self,
                                  yml_file: str,
                                  sandbox: Optional[bool] = True,
                                  git_repo: Optional[str] = None) -> DataSource:
        """
        Publishes metadata about Features to Pyrasgo

        :param yml_file: Rasgo compliant yml file that describes the feature(s) being created
        :param sandbox: Status of the features (True = 'Sandboxed' | False = 'Productionized')
        :param git_repo: Filepath string to these feature recipes in git
        :return: description of the features created
        """
        return self.publish.features_from_yml(yml_file, sandbox, git_repo)

    @deprecated_with_replacement('v0.3', 'publish.source_data')
    def publish_source_data(self,
                            source_type: str,
                            file_path: Optional[Path] = None,
                            df: Optional[pd.DataFrame] = None,
                            table: Optional[str] = None,
                            table_database:Optional[str] = None,
                            table_schema: Optional[str] = None,
                            data_source_name: Optional[str] = None,
                            data_source_domain: Optional[str] = None,
                            data_source_table_name: Optional[str] = None, parent_data_source_id: Optional[int] = None,
                            if_exists: Optional[str] = 'fail'
                            ) -> DataSource:
        """
        Push a csv, Dataframe, or table to a Snowflake table and register it as a Rasgo DataSource (TM)

        NOTES: csv files will import all columns as strings

        params:
            source_type: Values: ['csv', 'dataframe', 'table']
            df: pandas DataFrame (only use when passing source_type = 'dataframe')
            file_path: full path to a file on your local machine (only use when passing source_type = 'csv')
            table: name of a valid Snowflake table in your Rasgo account (only use when passing source_type = 'table')
            table_database: Optional: name of the database of the table passed in 'table' param (only use when passing source_type = 'table')
            table_schema: Optional: name of the schema of the table passed in 'table' param (only use when passing source_type = 'table')

            data_source_name: Optional name for the DataSource (if not provided a random string will be used)
            data_source_table_name: Optional name for the DataSource table in Snowflake (if not provided a random string will be used)
            data_source_domain: Optional domain for the DataSource (default is NULL)
            parent_data_source_id: Optional ID of a valid Rasgo DataSource that is a parent to this DataSource (default is NULL)

            if_exists: Values: ['fail', 'append', 'replace'] directs the function what to do if a DataSource already exists with this table name  (defaults to fail)

        return:
            Rasgo DataSource
        """
        return self.publish.source_data(source_type, file_path, df, table, table_database, table_schema, data_source_name, data_source_domain, data_source_table_name, parent_data_source_id, if_exists)

    @deprecated_with_replacement('v0.3', 'read.collection_data')
    def read_collection_data(self,
                             collection_id: int,
                             filters: Optional[Dict[str, str]] = None,
                             limit: Optional[int] = None) -> pd.DataFrame:
        """
        Constructs a pandas DataFrame from the specified Rasgo Collection

        :param collection_id: int
        :param filters: dictionary providing columns as keys and the filtering values as values.
        :param limit: integer limit for number of rows returned
        :return: Dataframe containing feature data
        """
        return self.read.collection_data(collection_id, filters, limit)

    @deprecated_with_replacement('v0.3', 'read.feature_data')
    def read_feature_data(self,
                          feature_id: int,
                          filters: Optional[Dict[str, str]] = None,
                          limit: Optional[int] = None) -> pd.DataFrame:
        """
        Constructs a pandas DataFrame from the specified Rasgo Feature data

        :param feature_id: int
        :param filters: dictionary providing columns as keys and the filtering values as values.
        :param limit: integer limit for number of rows returned
        :return: Dataframe containing feature data
        """
        return self.read.feature_data(feature_id, filters, limit)

    @deprecated_with_replacement('v0.3', 'read.source_data')
    def read_source_data(self,
                         source_id: int,
                         filters: Optional[Dict[str, str]] = None,
                         limit: Optional[int] = None) -> pd.DataFrame:
        """
        Constructs a pandas DataFrame from the specified Rasgo DataSource

        :param source_id: int
        :param filters: dictionary providing columns as keys and the filtering values as values.
        :param limit: integer limit for number of rows returned
        :return: Dataframe containing feature data
        """
        return self.read.source_data(source_id, filters, limit)


from typing import Dict, List, Optional, Union

from pyrasgo import primitives, schemas
from requests.exceptions import HTTPError

from .error import APIError


class Create():

    def __init__(self):
        from pyrasgo.config import get_session_api_key
        from .connection import Connection

        api_key = get_session_api_key()
        self.api = Connection(api_key=api_key)

    def collection(self, name: str,
                   type: Union[str, schemas.ModelType],
                   granularity: Union[str, schemas.Granularity],
                   description: Optional[str] = None,
                   is_shared: Optional[bool] = False) -> primitives.Collection:
        try:
            # If not enum, convert to enum first.
            model_type = type.name
        except AttributeError:
            model_type = schemas.ModelType(type)

        try:
            # If not enum, convert to enum first.
            granularity = granularity.name
        except AttributeError:
            granularity = schemas.Granularity(granularity)

        content = {"name": name,
                   "type": model_type.value,
                   "granularities": [{"name": granularity.value}],
                   "isShared": is_shared
                   }
        if description:
            content["description"] = description
        response = self.api._post("/models", _json=content, api_version=1)
        return primitives.Collection(api_object=response.json())

    def data_source(self, table: str,
                    name: str,
                    source_type: str,
                    database: Optional[str] = None,
                    schema: Optional[str] = None,
                    source_code: Optional[str] = None,
                    domain: Optional[str] = None,
                    parent_source_id: Optional[int] = None) -> primitives.DataSource:
        data_source = schemas.DataSourceCreate(name=name,
                                               table=table,
                                               tableDatabase=database,
                                               tableSchema=schema,
                                               sourceCode=source_code,
                                               domain=domain,
                                               sourceType=source_type,
                                               parentId=parent_source_id)
        response = self.api._post("/data-source", data_source.dict(exclude_unset=True), api_version=1).json()
        return primitives.DataSource(api_object=response)

    def dataframe(self, unique_id: str,
                  name: str = None,
                  shared_status: str = 'organization',
                  column_hash: Optional[str] = None,
                  update_date: str = None) -> schemas.Dataframe:
        shared_status = 'organization' if shared_status not in ['public', 'private'] else shared_status
        dataframe = schemas.DataframeCreate(uniqueId=unique_id,
                                            name=name,
                                            sharedStatus=shared_status,
                                            columnHash=column_hash,
                                            updatedDate=update_date)
        try:
            response = self.api._post("/dataframes", dataframe.dict(exclude_unset=True), api_version=1).json()
        except HTTPError as e:
            error_message = f"Failed to create dataframe {unique_id}."
            if e.response.status_code == 409:
                error_message += f" This id is already in use in your organization. Dataframe IDs must be unique."
            raise APIError(error_message)
        return schemas.Dataframe(**response)

    def data_source_stats(self, data_source_id: int):
        """
        Sends an api request to build stats for a specified data source.
        """
        return self.api._post(f"/data-source/profile/{data_source_id}", api_version=1).json()

    def data_source_feature_stats(self, data_source_id: int):
        """
        Sends an api request to build stats for all features in a specified data source.
        """
        return self.api._post(f"/data-source/{data_source_id}/features/stats", api_version=1).json()

    def feature(self,
                data_source_id: int,
                display_name: str,
                column_name: str,
                description: str,
                # data_source_column_id: int,
                status: str,
                git_repo: str,
                tags: Optional[List[str]] = None) -> primitives.Feature:
        feature = schemas.FeatureCreate(name=display_name,
                                        code=column_name,
                                        description=description,
                                        dataSourceId=data_source_id,
                                        orchestrationStatus=status,
                                        tags=tags or [],
                                        gitRepo=git_repo)
        try:
            response = self.api._post("/features/", feature.dict(exclude_unset=True), api_version=1).json()
        except HTTPError as e:
            error_message = f"Failed to create Feature {display_name}."
            if e.response.status_code == 409:
                error_message += f" {column_name} already has a feature associated with it. Try running update feature instead."
            raise APIError(error_message)
        return primitives.Feature(api_object=response)

    def feature_stats(self, feature_id: int) -> Dict:
        """
        Sends an api request to build feature stats for a specified feature.
        """
        return self.api._post(f"/features/{feature_id}/stats", api_version=1).json()

    def feature_importance_stats(self, id: int, payload: schemas.FeatureImportanceStats) -> Dict:
        """
        Sends an api requrest to build feature importance stats for the specified model
        """
        return self.api._post(f"/models/{id}/stats/feature-importance", payload.dict(), api_version=1).json()

    def column_importance_stats(self, id: str, payload: schemas.FeatureImportanceStats) -> Dict:
        """
        Sends a json payload of importance from a dataFrame to the API so it can render in the WebApp
        """
        return self.api._post(f"/dataframes/{id}/feature-importance", payload.dict(), api_version=1).json()

    def dataframe_profile(self, id: str, payload: schemas.ColumnProfiles) -> Dict:
        """
        Send a json payload of a dataframe profile so it can render in the WebApp
        """
        return self.api._post(f"/dataframes/{id}/profile", payload.dict(), api_version=1).json()

    def transform(
        self,
        *,
        name: str,
        source_code: str,
        type: Optional[str] = None,
        arguments: Optional[List[dict]] = None,
        description: Optional[str] = None
    ) -> schemas.Transform:
        """
        Create and return a new Transform in Rasgo
        Args:
            name: Name of the Transform
            source_code: Source code of transform
            type: Type of transform it is. Used for categorization only
            arguments: A list of arguments to supply to the transform
                       so it can render them in the UI. Each argument
                       must be a dict with the keys: 'name', 'description', and 'type'
                       values all strings for their corresponding value
            description: Description of Transform

        Returns:
            Created Transform obj
        """
        arguments = arguments if arguments else []

        transform = schemas.TransformCreate(
            name=name,
            type=type,
            sourceCode=source_code,
            description=description
        )
        transform.arguments = [
            schemas.TransformArgumentCreate(**x) for x in arguments
        ]
        response = self.api._post("/transform", transform.dict(), api_version=1).json()
        return schemas.Transform(**response)

    # ----------------------------------
    #  Internal/Private Create Calls
    # ----------------------------------

    def _dataset(
            self,
            *,
            name: str,
            resource_key: Optional[str] = None,
            description: Optional[str] = None,
            status: Optional[str] = 'published',
            dw_table_id: Optional[int] = None,
            dw_operation_set_id: Optional[int] = None,
            fqtn: Optional[str] = None,
            attributes: Optional[dict] = None,
            table_type: Optional[str] = "VIEW"
    ) -> schemas.Dataset:
        """
        Create a Dataset in Rasgo.

        Args:
            name: Name of the dataset
            description: Description of the dataset
            status: Status of whether this datasets is published or not
            dw_table_id: DW table to associate with this Dataset
            dw_operation_set_id: Id of the Operation Set to associate with this Dataset
            fqtn: Fully qualified table name of the table to register this Dataset as
            attributes: Dictionary containing dataset attributes to be published
            table_type: Type of object to create in snowflake. Can be "TABLE" or "VIEW"
        Returns:
            Created Dataset Obj
        """
        if (not table_type or table_type.upper() not in ("TABLE", "VIEW")):
            raise ValueError(f"table_type {table_type} is not usable. Please make sure you select either 'TABLE' or 'VIEW'.")

        dataset_create = schemas.DatasetCreate(
            name=name,
            resource_key=resource_key,
            description=description,
            status=status,
            dw_table_id=dw_table_id,
            dw_operation_set_id=dw_operation_set_id,
            attributes=attributes
        )
        path = f"/datasets"
        if fqtn:
            path += f"?fqtn={fqtn}"
        else:
            path += f"?table_type={table_type}"
        response = self.api._post(
            path, dataset_create.dict(), api_version=2
        ).json()
        return schemas.Dataset(**response)

    def _operation_set(
            self,
            operations: List[schemas.OperationCreate],
            dataset_dependency_ids: List[int]
    ) -> schemas.OperationSet:
        """
        Create a operation set in Rasgo with specified operation
        and input dataset dependencies ids
        """
        operation_set_create = schemas.OperationSetCreate(
            operations=operations,
            dataset_dependency_ids=dataset_dependency_ids
        )
        response = self.api._post(
            "/operation-sets", operation_set_create.dict(), api_version=2
        ).json()
        return schemas.OperationSet(**response)

    def _operation_set_async(
            self,
            operations: List[schemas.OperationCreate],
            dataset_dependency_ids: List[int]
    ) -> schemas.OperationSetAsyncTask:
        """
        Create a operation set in Rasgo with specified operation
        and input dataset dependencies ids
        """
        operation_set_create = schemas.OperationSetCreate(
            operations=operations,
            dataset_dependency_ids=dataset_dependency_ids
        )
        response = self.api._post(
            "/operation-sets/async", operation_set_create.dict(), api_version=2
        ).json()
        return schemas.OperationSetAsyncTask(**response)

    def _operation_set_preview(
            self,
            operations: List[schemas.OperationCreate],
            dataset_dependency_ids: List[int]
    ) -> str:
        """
        Create a operation set in Rasgo with specified operation
        and input dataset dependencies ids
        """
        operation_set_create = schemas.OperationSetCreate(
            operations=operations,
            dataset_dependency_ids=dataset_dependency_ids
        )
        response = self.api._post(
            "/operation-sets/offline", operation_set_create.dict(), api_version=2
        ).json()
        return response

    def _operation_set_preview_async(
            self,
            operations: List[schemas.OperationCreate],
            dataset_dependency_ids: List[int]
    ) -> schemas.OperationSetOfflineAsyncTask:
        """
        Create a operation set in Rasgo with specified operation
        and input dataset dependencies ids in an async fashion
        """
        operation_set_create = schemas.OperationSetCreate(
            operations=operations,
            dataset_dependency_ids=dataset_dependency_ids
        )
        response = self.api._post(
            "/operation-sets/offline/async", operation_set_create.dict(), api_version=2
        ).json()
        return schemas.OperationSetOfflineAsyncTask(**response)

    def _dataset_correlation_stats(self, *, table_id: int, only_if_data_changed: Optional[bool] = True) -> None:
        """
        Trigger stats generation on a dataset
        """

        # Note: dimension_column_id should not be passed, as it has already been set at publish time
        stats_create = schemas.GenerateStat(
            dwTableId=table_id,
            dimensionColumnId=None,
            onlyIfDataChanged=only_if_data_changed
        )
        self.api._post("/stats", stats_create.dict(), api_version=2)

    def _operation_render(
            self,
            operation: schemas.OperationCreate
    ) -> str:
        """
        Test the rendering of an operation
        """
        response = self.api._post(
            "/operation/render", operation.dict(), api_version=2
        ).json()
        return response



from datetime import datetime
from typing import Dict, List, Optional, Any

from pydantic import BaseModel


class TransformArgumentCreate(BaseModel):
    name: str
    description: str
    type: str
    is_optional: Optional[bool]


class TransformArgument(TransformArgumentCreate):
    pass


class TransformCreate(BaseModel):
    name: Optional[str]
    type: Optional[str]
    sourceCode: str
    description: Optional[str]
    arguments: Optional[List[TransformArgumentCreate]]


class TransformUpdate(BaseModel):
    """
    Contract for updating a Transform
    """
    name: Optional[str]
    type: Optional[str]
    description: Optional[str]
    sourceCode: Optional[str]
    arguments: Optional[List[TransformArgumentCreate]]


class Transform(TransformCreate):
    id: Optional[int]
    arguments: Optional[List[TransformArgument]]


# Transform Execution
class TransformArgs(BaseModel):
    """
    Schema/contract for Executing a Transform
    """
    transformId: Optional[int]
    sourceCode: Optional[str]
    arguments: Dict[str, Any]


class TransformExecute(BaseModel):
    """
    Contract for execution of one or more transforms
    """
    dataSourceId: int
    newTableName: Optional[str]
    transformArgs: List[TransformArgs]



"""
Module for all things Source Transforms
"""
from typing import Dict, List, Optional, Union

import pandas as pd

from pyrasgo.api.connection import Connection
from pyrasgo.api.error import APIError
from pyrasgo.primitives.source import DataSource
from pyrasgo.schemas import transform as schemas


class TransformChain(Connection):
    """
    Class for implementing executing, previewing, and source creation of one
    or more transforms
    """

    def __init__(self,
                 source: DataSource,
                 transform_steps: Optional[List[schemas.TransformArgs]] = None,
                 **kwargs) -> None:
        super().__init__(**kwargs)
        from pyrasgo.api import Get
        from pyrasgo.storage import DataWarehouse, SnowflakeDataWarehouse

        self.transform_steps = transform_steps if transform_steps else []
        self.source = source
        self._get_api = Get()
        self._data_warehouse: SnowflakeDataWarehouse = DataWarehouse.connect()

    @property
    def has_any_non_db_transforms(self) -> bool:
        """
        Return True if any of the supplied transform steps in the Chain
        are not from the db, but from source code.
        """
        return any([
            not bool(transform_step.transformId)
            for transform_step in self.transform_steps
        ])

    def transform(self,
                  *,
                  transform_id: Optional[int] = None,
                  transform_name: Optional[str] = None,
                  transform: Optional[schemas.Transform] = None,
                  source_code: Optional[str] = None,
                  arguments: Optional[Dict[str, str]] = None,
                  **kwargs: Union[str, int, List, Dict]) -> 'TransformChain':
        """
        Add/chain another transform step on this source, and return self/TransformerChain obj

        Returned obj can chain more transformers together by calling .transform() again if needed.

        Preview the transformed dataset as a df or SQL as a df with .preview() and .preview_sql()

        Create a source from the chained transforms by then calling .to_source(). You can ONLY do this
        when supplying a transform id, name, or obj and not source_code

        Args:
            transform_id: Transform id to add as step to chain. Can input Transform name or Transform schema instead
            transform_name: Transform name to add as step to chain. Can add Transform id or Transform schema instead
            transform: Transform schema to add as step to chain. Can be Transform id or name instead
            source_code: If not entering a transform id, name, or obj, test source_code with the transform
            arguments: Arguments to apply to this transform

        Returns:
            A TransformChain() obj with a newly added transform step
        """
        arguments = arguments if arguments else {}

        # Get and validate transform obj input
        self._validate_transform_inputs(
            transform_id, transform_name, transform, source_code
        )
        transform = self._get_transform_model(
            transform_id, transform_name, transform, source_code
        )

        # Update the Transform arguments with any supplied kwargs
        arguments.update(kwargs)

        # Create a new instance of a TransformChain
        # class with same source and steps
        new_transform_chain = self.__class__(
            source=self.source,
            transform_steps=self.transform_steps
        )

        # Append Transform Args as new step in new Chain Obj
        # and return new TransformChain
        transform_step = schemas.TransformArgs(
            transformId=transform.id if hasattr(transform, 'id') else None,
            sourceCode=transform.sourceCode,
            arguments=arguments,
        )
        new_transform_chain.transform_steps.append(transform_step)
        return new_transform_chain

    def preview_sql(self) -> str:
        """
        Preview the generated SQL for this Transform Chain
        """
        self._raise_error_if_no_transform_steps()

        transform_execute = schemas.TransformExecute(
            dataSourceId=self.source.id,
            transformArgs=self.transform_steps,
        )
        transform_sql = self._post(resource="/transform/sql",
                                   _json=transform_execute.dict(exclude_unset=True),
                                   api_version=1).json()
        return transform_sql

    def preview(self, limit: int = 10) -> Union[pd.DataFrame, str]:
        """
        Preview the dataset from this chain of transforms

        Returns a pandas DataFrame

        Args:
            limit: Number of rows to return; default 10

        Returns:
            DataFrame of the Transform Chained source if no errors
        """
        self._raise_error_if_no_transform_steps()

        # Get the transform SQL and set limit
        transform_sql = self.preview_sql()
        if limit:
            transform_sql = f"{transform_sql} LIMIT {limit}"

        # Execute the Transform Chain SQL query. Show error message if any
        try:
            df = self._data_warehouse.query_into_dataframe(transform_sql)
            return df
        except Exception as e:
            raise APIError(f"Error Executing transform SQL\n\nerror\n-----\n{e} "
                           f"{e}\n\ngenerated_sql\n--------------\n{transform_sql}")

    def to_source(self, new_table_name: Optional[str] = None) -> DataSource:
        """
        Create a new DataSource for this Chain of Transforms
        """
        self._raise_error_if_no_transform_steps()

        # Don't allow creating source if any transform steps
        # are from source_code and not DB
        if self.has_any_non_db_transforms:
            raise APIError("One or more of your transforms were created "
                           "from Jinja source_code and not the db. To call the .to_source() "
                           "method, all transforms need to be created already.\n\nCall "
                           "rasgo.get.transforms() and use any of those available. "
                           "Create a new Transform with rasgo.create.transform()")

        # If new source name entered make sure no spaces in name
        if new_table_name and " " in new_table_name:
            raise ValueError(f"The new source is going to be a table. "
                             f"Please pass a name without whitespace")

        # Execute the transforms on this source to create a new one
        transform_execute = schemas.TransformExecute(
            dataSourceId=self.source.id,
            transformArgs=self.transform_steps,
            newTableName=new_table_name
        )
        response = self._post(resource="/transform/execute",
                              _json=transform_execute.dict(exclude_unset=True),
                              api_version=1).json()
        return DataSource(api_object=response)

    @staticmethod
    def _validate_transform_inputs(transform_id: Optional[int] = None,
                                   transform_name: Optional[str] = None,
                                   transform: Optional[schemas.Transform] = None,
                                   source_code: Optional[str] = None) -> None:
        """
        Validate the user only inputted one transform (id, name, or obj)

        Also validate that a transform with id or name exists if one of those were entered
        """
        # Make sure source_code and not transform identifier supplied
        if source_code and (transform_id or transform_name or transform):
            raise APIError("You cannot execute a transform by supplying a Transform "
                           "identifier(id, name, or obj) and 'source_code'. "
                           "Please supply one or the other.")

        # Make sure only one transform identifier entered or source code not entered
        if not source_code and not transform_id and not transform and not transform_name:
            raise APIError("To execute a Transform supply the param "
                           "'transform_id' (id of transform), 'transform_name' "
                           "(name of transform), 'transform' (Transform obj), "
                           "or 'source_code' (Transform Jinja source code to apply)"
                           "from rasgo.get.transforms()")

        elif transform_id and (transform_name or transform) \
                or transform_name and (transform_id or transform) \
                or transform and (transform_id or transform_name):
            raise APIError("Please supply ONLY one of 'transform_id', 'transform_name' or 'transform' obj "
                           "as input params to identify the transform to use")

    def _get_transform_model(self,
                             transform_id: Optional[int] = None,
                             transform_name: Optional[str] = None,
                             transform: Optional[schemas.Transform] = None,
                             source_code: Optional[str] = None) -> schemas.Transform:
        """
        Get and return the transform model by id, name, or source code
        """
        # Validate transform with id or name exists if entered
        if transform_id:
            transform = self._get_api.transform(transform_id)
            if not transform:
                raise ValueError(f"No Transform found with id {transform_id}")
        elif transform_name:
            transform = self._get_transform_by_name(transform_name)
            if not transform:
                raise ValueError(f"No Transform found with name {transform_name}")

        # If source code provided init model with source code
        # We aren't adding arguments as their only used here for validation,
        # we don't have have jinja installed on the SDK to parse it,
        # and on the API side an error will be raised if any missing during execution
        elif source_code:
            transform = schemas.Transform(sourceCode=source_code,
                                          arguments=[])
        return transform

    def _get_transform_by_name(self, transform_name: str) -> Optional[schemas.Transform]:
        """
        Get and return a transform obj by name
        """
        available_transforms = self._get_api.transforms()
        for transform in available_transforms:
            if transform_name == transform.name:
                return transform

    def _raise_error_if_no_transform_steps(self) -> None:
        if not self.transform_steps:
            raise APIError(
                "No transform steps have been added to this chain yet. call .transform(**kwargs) "
                "first to preview the transform chain or create a source from it."
            )



from enum import Enum
import pandas as pd
from typing import List, Optional

from pyrasgo.api.error import APIError
from pyrasgo.utils import naming


def build_schema(df: pd.DataFrame, include_index=False) -> dict:
    """
    Returns a dict representing the schema of a dataframe
    """
    from pandas.io.json import build_table_schema
    schema_list = build_table_schema(df)
    if not include_index:
        return {column['name']: column
                for column in schema_list['fields'] if column['name'] != 'index'}
    return {column['name']: column
            for column in schema_list['fields']}


def generate_ddl(df: pd.DataFrame,
                 table_name: str,
                 append: Optional[bool] = False):
    """
    Generates a SQL statement to create a table matching the schema of a dataframe
    """
    #default is overwrite
    create_statement = "CREATE TABLE IF NOT EXISTS" if append else "CREATE OR REPLACE TABLE"
    return pd.io.sql.get_schema(df, table_name) \
        .replace("CREATE TABLE", create_statement) \
        .replace('"', '')

def confirm_df_columns(df: pd.DataFrame, dimensions: List[str], features: List[str]=[]):
    """
    Checks whether a list of dimensions and features are in a dataframe
    then returns valid columns
    """
    return confirm_list_columns(list(df.columns), dimensions, features)

def confirm_list_columns(columns: list, dimensions: List[str], features: List[str]=[]):
    """
    Checks whether a list of dimensions and features are in a list of columns
    then returns valid columns
    """
    missing_dims = []
    missing_features = []
    valid_dims = []
    valid_features = []
    consider = []
    for dim in dimensions:
        if dim not in columns:
            missing_dims.append(dim)
            if naming._cleanse_sql_name(dim) in columns:
                consider.append(naming._cleanse_sql_name(dim))
        else:
            valid_dims.append(dim)
    if len(features) == 0:
        features = set(columns) - set(valid_dims)
    for ft in list(features):
        if ft not in columns:
            missing_features.append(ft)
            if naming._cleanse_sql_name(ft) in columns:
                consider.append(naming._cleanse_sql_name(ft))
        else:
            valid_features.append(ft)
    if missing_dims or missing_features:
        raise APIError(f"Specified columns do not exist in dataframe: "
                        f"Dimensions({missing_dims}) Features({missing_features}) "
                        f"Consider these: ({consider})?")
    return valid_dims, valid_features

def generate_unique_id() -> str:
    """
    Calculates a unique 32 char string that can be used in secure urls
    """
    import secrets
    return secrets.token_urlsafe(32)

def map_pandas_df_type(pandas_df_type: str) -> str:
    """For a given pandas dataframe type, return the equivalent python type.
    Default to input if no match"""
    if pandas_df_type.startswith('float'):
        return 'float'
    if pandas_df_type.startswith('datetime'):
        return 'datetime'
    if pandas_df_type.startswith('int'):
        return 'integer'
    return pandas_df_type

def set_df_id(df: pd.DataFrame, experiment_id: Optional[str]) -> int:
    """
    Checks whether a df is part of an ongoing experiment and tags it with
    the experiment id or a new id
    """
    # If ID is set via experiment, use it
    if experiment_id:
        tag_dataframe(df, {"RasgoID": experiment_id})
        return experiment_id
    # If df is already marked with an ID, keep it
    elif df.attrs.get("RasgoID"):
        return df.attrs["RasgoID"]
    # Else calcualte a new ID
    else:
        unique_id = generate_unique_id()
        tag_dataframe(df, {"RasgoID": unique_id})
        return unique_id

def tag_dataframe(df: pd.DataFrame, attribute: dict):
    """
    Tags a dataframe with a user-defined attribute
    """
    write_back_attrs = df.attrs or {}
    write_back_attrs.update(attribute)
    df.attrs = write_back_attrs

def _cleanse_sql_dataframe(df: pd.DataFrame):
    """
    Renames all columns in a pandas dataframe to Snowflake compliant names in place
    """
    df.rename(columns={r: naming._cleanse_sql_name(r) for r in build_schema(df)},
                inplace=True)
#TODO: currently the names in this enum are string representations of pandas
# dataframe types. Can we make the names actual type objects? If so, we could
# remove a lot of random error handling going on in evaluate.py
class DataframeDataTypes(Enum):
    INT64 = "number"
    INT32 = "number"
    INT16 = "number"
    INT8 = "number"
    FLOAT64 = "number"
    FLOAT32 = "number"
    FLOAT16 = "number"
    FLOAT8 = "number"
    UINT64 = "number"
    UINT32 = "number"
    UINT16 = "number"
    UINT8 = "number"
    DATETIME64 = "string"
    OBJECT = "string"
    STRING = "string"
    BOOL = "boolean"



from pydantic import BaseModel, Field
from typing import Optional


class DatasetColumn(BaseModel):
    """
    Contract to return from get by id endpoints
    """
    id: int
    name: str = Field(alias='columnName')
    display_name: Optional[str] = Field(alias='displayName')
    data_type: str = Field(alias='dataType')
    description: Optional[str]
    is_feature: bool = Field(alias='isFeature')
    is_dimension: bool = Field(alias='isDimension')
    dw_column_id: int = Field(alias='dwColumnId')

    class Config:
        allow_population_by_field_name = True


class DatasetColumnCreate(BaseModel):
    """
    Contract to accept in post endpoints
    """
    name: str = Field(alias='columnName')
    display_name: Optional[str] = Field(alias='displayName')
    data_type: str = Field(alias='dataType')
    description: Optional[str]
    is_feature: bool = Field(alias='isFeature')
    is_dimension: bool = Field(alias='isDimension')
    dw_column_id: int = Field(alias='dwColumnId')

    class Config:
        allow_population_by_field_name = True


class DatasetColumnUpdate(BaseModel):
    """
    Contract to accept in put endpoints
    """
    name: Optional[str] = Field(alias='columnName')
    display_name: Optional[str] = Field(alias='displayName')
    data_type: Optional[str] = Field(alias='dataType')
    description: Optional[str]
    is_feature: Optional[bool] = Field(alias='isFeature')
    is_dimension: Optional[bool] = Field(alias='isDimension')
    dw_column_id: Optional[int] = Field(alias='dwColumnId')

    class Config:
        allow_population_by_field_name = True


class DatasetColumnUpdateById(DatasetColumnUpdate):
    """
    Contract to accept in put endpoints

    BMS:
    This is a bit of a dice-roll around our framework.
    The DatasetColumnUpdate contract is used in the PUT /v2/dataset-columns/{dataset_column_id} endpoint.
    Since the id is in the path, we don't need the id as part of the contract.

    This contract is used in the PUT /datasets/{dataset_id} endpoint. That endpoint accepts a DatasetUpdate contract,
    which in turn accepts a list of this contract. This contract is identical to its parent, except that we ask the user
    to provide the ids since we don't have them in the path parameter.

    This is a departure from our normal pattern of using match() to try to find the identity of an object in lieu of a
    user directly telling us.

    If this becomes a pain, we'll rip it out and revert back to using match().
    """
    id: int = Field()

    class Config:
        allow_population_by_field_name = True



import pandas as pd
from typing import Dict, List, Optional

from .error import APIError
from pyrasgo import config
from pyrasgo import schemas as api
from pyrasgo.primitives import Collection, Feature, FeatureList, DataSource, Dataset


class Read():

    def __init__(self):
        from . import Get
        from pyrasgo.storage import DataWarehouse, SnowflakeDataWarehouse

        self.data_warehouse: SnowflakeDataWarehouse = DataWarehouse.connect()
        self.get = Get()

    def collection_data(self,
                        id: int,
                        filters: Optional[Dict[str, str]] = None,
                        limit: Optional[int] = None) -> pd.DataFrame:
        """
        Constructs a pandas DataFrame from the specified Rasgo Collection

        :param id: int
        :param filters: dictionary providing columns as keys and the filtering values as values.
        :param limit: integer limit for number of rows returned
        :return: Dataframe containing feature data
        """
        collection = self.get.collection(id)
        if collection:
            try:
                table_metadata = collection._make_table_metadata()
                query, values = self.data_warehouse.make_select_statement_filter_dict(table_metadata, filters, limit)
                return self.data_warehouse.query_into_dataframe(query, values)
            except Exception as e:
                raise APIError(f"Collection table is not reachable: {e}")
        raise APIError("Collection does not exist")

    def feature_data(self,
                     id: int,
                     filters: Optional[Dict[str, str]] = None,
                     limit: Optional[int] = None) -> pd.DataFrame:
        """
        Constructs a pandas DataFrame from the specified Rasgo Feature data

        :param id: int
        :param filters: dictionary providing columns as keys and the filtering values as values.
        :param limit: integer limit for number of rows returned
        :return: Dataframe containing feature data
        """
        feature = self.get.feature(id)
        if feature.sourceTable:
            try:
                table_metadata = feature._make_table_metadata()
                #TODO: if we ever support multiple features, add them to this line -
                features = feature.columnName
                indices = ','.join(feature.indexFields)
                columns = indices +', '+features
                query, values = self.data_warehouse.make_select_statement_filter_dict(table_metadata, filters, limit, columns)
                return self.data_warehouse.query_into_dataframe(query, values)
            except Exception as e:
                raise APIError(f"Feature table is not reachable: {e}")
        raise APIError("Feature table does not exist")

    def source_data(self,
                    id: int,
                    filters: Optional[Dict[str, str]] = None,
                    limit: Optional[int] = None) -> pd.DataFrame:
        """
        Constructs a pandas DataFrame from the specified Rasgo DataSource

        :param id: int
        :param filters: dictionary providing columns as keys and the filtering values as values.
        :param limit: integer limit for number of rows returned
        :return: Dataframe containing feature data
        """
        data_source = self.get.data_source(id)
        if data_source:
            try:
                table_metadata = data_source._make_table_metadata()
                query, values = self.data_warehouse.make_select_statement_filter_dict(table_metadata, filters, limit)
                return self.data_warehouse.query_into_dataframe(query, values)
            except Exception as e:
                raise APIError(f"DataSource table is not reachable: {e}")
        raise APIError("DataSource does not exist")

    def dataset(
            self,
            id: Optional[int] = None,
            dataset: Optional[Dataset] = None,
            filters: Optional[List[str]] = None,
            order_by: Optional[List[str]] = None,
            columns: Optional[List[str]] = None,
            limit: Optional[int] = None,
            snapshot_index: Optional[int] = None
    ) -> pd.DataFrame:
        """
        Constructs and returns pandas DataFrame from the specified Rasgo Dataset

        You can supply SQL WHERE clause filters, order the dataset by columns, only
        return selected columns, and add a return limit as well

        Example:
            ```
            rasgo.read.dataset(
                id=74,
                filters=['SALESTERRITORYKEY = 1', 'TOTALPRODUCTCOST BETWEEN 1000 AND 2000'],
                order_by=['TOTALPRODUCTCOST'],
                columns=['PRODUCTKEY', 'TOTALPRODUCTCOST', 'SALESTERRITORYKEY'],
                limit=50
            )
            ```

        Args:
            id: dataset id to read into df
            dataset: Dataset obj to read into df
            filters: List of SQL WHERE filters strings to filter on returned df
            order_by: List of columns to order by in returned dataset
            columns: List of columns to return in the df
            limit: Only return this many rows in the df
            snapshot: the index of a snapshot from Dataset.snapshots to read 
        """
        # Validate one dataset passed in, id or dataset obj
        if not dataset and id is None:
            raise ValueError(f"Must pass either a valid dataset ID or Dataset "
                             f"object to read into a DataFrame")
        if not dataset:
            # Note: Func below already raises API error if dataset with id doesn't exist
            dataset = self.get.dataset(id)
        
        fqtn = dataset.fqtn

        if snapshot_index:
            try:    
                fqtn = dataset._api_dataset.snapshots[snapshot_index].fqtn
            except IndexError:
                raise ValueError(f"Snapshot index does not exist")

        if dataset:
            try:
                query = self.data_warehouse.make_select_statement(
                    table_metadata={'fqtn': fqtn},
                    filters=filters,
                    order_by=order_by,
                    columns=columns,
                    limit=limit
                )
                return self.data_warehouse.query_into_dataframe(query)
            except Exception as e:
                raise APIError(f"Dataset table is not reachable: {e}")




from types import FunctionType
import inspect
import pandas
import typing

def example_udf(df: pandas.DataFrame) -> pandas.DataFrame:
    # transform dataframe
    print('test')
    return df

def is_valid_udf(udf: FunctionType): 
    param_types = list(inspect.signature(udf).parameters.values())
    if len(param_types) != 1 or not param_types[0].annotation is pandas.DataFrame:
        raise TypeError(f"Valid User Defined Functions must accept exactly one parameter of explicit type<pandas.DataFrame>\n\n e.g\n\n {inspect.getsource(example_udf)}")

    if not inspect.signature(udf).return_annotation is pandas.DataFrame: 
        raise TypeError(f"Valid User Defined Functions must have a return type of exactly one <pandas.DataFrame>\n\n e.g\n\n {inspect.getsource(example_udf)}")

def stringify_udf(udf: FunctionType) -> str:
    """Creates a string representation of the given udf.  Aliases the udf as 'udf' 
    since we don't know the name of the function"""
    return f'{inspect.getsource(example_udf)}' \
        f'\nudf = {example_udf.__name__}' 
        
def _udf_text_to_function(function_text: str, dataframe: pandas.DataFrame) -> pandas.DataFrame:
    """  pass reference to global symbols so function can use pandas, builtins, etc. 
    TODO: allow users to bring their own imports, add them to passed globals 
    TODO: pass a new globals object so users can't override our symbols
    This statement executes the function text and adds the defined function to
    the locs dict. We then return that function. 
    """
    locs = {}
    exec(function_text, globals(), locs)
    return locs['udf']





"""
Dataset 'Primitive' In rasgo SDK
"""
import functools
import inspect
from inspect import Parameter
from datetime import datetime
from typing import Callable, Dict, List, Optional, Union

import pandas as pd
from pyrasgo import schemas
from pyrasgo.api.connection import Connection
from pyrasgo.api.error import APIError
from pyrasgo.schemas.dataset import DatasetBulk
from pyrasgo.utils import naming, polling

# Value of dataset.status if dataset is published
DS_PUBLISHED_STATUS = 'published'

def require_operation_set(func: Callable) -> Callable:
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        self: 'Dataset' = args[0]
        self._require_operation_set()
        return func(*args, **kwargs)
    return wrapper


class Dataset(Connection):
    """
    Representation of a Rasgo Dataset
    """
    def __init__(self,
                 # Args passed from Rasgo
                 api_dataset: Optional[schemas.Dataset] = None,
                 api_operation_set: Optional[schemas.OperationSet] = None,
                 # Args passed from transforming
                 operations: Optional[List[schemas.OperationCreate]] = None,
                 dataset_dependencies: Optional[List[int]] = None,
                 guid: Optional[str] = None,
                 transforms: Optional[List[schemas.Transform]] = None,
                 verbose=False,
                 async_compute: Optional[bool] = True, # TODO: Remove before next major version
                 **kwargs: Dict):
        """
        Init functions in two modes:
            1. This Dataset retrieved from Rasgo. This object is for reference, and cannot
               be changed, but can be transformed to build new datasets
            2. This Dataset represents a new dataset under construction. It is not persisted in Rasgo
               and instead consists of some operations that will be used to generate a new dataset.
        """
        super().__init__(**kwargs)

        self._verbose = verbose
        self._api_dataset: schemas.Dataset = api_dataset
        self._api_operation_set: schemas.OperationSet = api_operation_set
        self._operations = operations if operations else []
        self._dataset_dependencies = dataset_dependencies if dataset_dependencies else []
        self._guid = guid
        self._source_code_preview = None
        if transforms:
            self._available_transforms = transforms
        else:
            self._available_transforms = _get_transforms()
        self._async_compute = async_compute

        #  alias .transform allowing direct referencing of named transforms
        for transform in self._available_transforms:
            f = self._create_udt_function(transform)
            setattr(self, transform.name, f)

    def __repr__(self) -> str:
        """
        Get string representation of this dataset
        """
        if self._api_dataset:
            return f"Dataset(id={self.id}, " \
                   f"name={self.name}, " \
                   f"resource_key={self._api_dataset.resource_key}, " \
                   f"version={self._api_dataset.version}, " \
                   f"status={self.status}, " \
                   f"description={self.description})"
        else:
            return f"Dataset()"

    # -------------------
    # Properties
    # -------------------

    @property
    def id(self) -> Optional[int]:
        """
        Return the id for this dataset

        Raise API error if one doesn't exist yet and is an offline dataset
        """
        if self._api_dataset:
            return self._api_dataset.id

    @property
    def name(self) -> Optional[str]:
        """
        Return dataset name if set/saved
        """
        if self._api_dataset:
            return self._api_dataset.name

    @property
    def description(self) -> Optional[str]:
        """
        Return dataset description if set/saved
        """
        if self._api_dataset:
            return self._api_dataset.description

    @property
    def status(self) -> Optional[str]:
        """
        Return dataset status
        """
        if self._api_dataset:
            return self._api_dataset.status

    @property
    @require_operation_set
    def fqtn(self) -> str:
        """
        Returns the Fully Qualified Table Name for this dataset
        """
        if self._api_dataset and self._api_dataset.dw_table:
            return self._api_dataset.dw_table.fqtn
        elif self._guid:
            dw_namespace = self._get_default_namespace()
            return f"{dw_namespace['database']}." \
                   f"{dw_namespace['schema']}." \
                   f"{self._guid}"
        else:
            raise AttributeError("No data warehouse fqtn exists for this Dataset")

    @property
    def columns(self) -> Optional[List[schemas.DatasetColumn]]:
        """
        Return the columns for this dataset if it is from the API
        """
        if self._api_dataset:
            return self._api_dataset.columns


    @property
    def created_date(self) -> Optional[datetime]:
        """
        Return date this dataset was created
        """
        if self._api_dataset:
            return self._api_dataset.create_timestamp

    @property
    def update_date(self) -> Optional[datetime]:
        """
        Return date this dataset was updated
        """
        if self._api_dataset:
            return self._api_dataset.update_timestamp

    @property
    def attributes(self) -> Optional[Dict]:
        """
        Return the attributes for this dataset if it is from the API
        """
        if self._api_dataset:
            return self._api_dataset.attributes

    @property
    def dependencies(self) -> List['Dataset']:
        """
        Return a list of dataset dependencies for this dataset
        """
        from pyrasgo.api import Get

        get = Get()
        dataset_deps = []
        self._cache_op_set_from_api()
        if self._api_operation_set:
            for ds in self._api_operation_set.dataset_dependencies:
                dataset_deps.append(get.dataset(ds.id))
        else:
            get = Get()
            for ds_id in self._dataset_dependencies:
                dataset_deps.append(get.dataset(ds_id))
        return dataset_deps


    @property
    @require_operation_set
    def sql(self) -> Optional[str]:
        """
        Return the source code SQL used to generate this dataset
        """
        # Raise error if no operations/source code for this dataset yet
        has_operations = (not hasattr(self._api_operation_set, "operations")
                          or (not self._api_operation_set.operations))
        if has_operations and not self._source_code_preview:
            return
        if self._api_dataset:
            return '\n\n'.join([x.operation_sql for x in self._api_operation_set.operations])
        else:
            return self._source_code_preview

    @property
    def snapshots(self) -> Optional[list]:
        """return a list of tuples of timestamp identified by index and the snapshot creation timestamp"""
        return [(i+1, ss.timestamp) for i, ss in enumerate(self._api_dataset.snapshots)]
    
    @property
    def versions(self) -> List[DatasetBulk]:
        """return a list of versions of this dataset"""
        return self._get(
                    f"datasets/rk/{self._api_dataset.resource_key}/versions",
                    api_version=2
                ).json()

    @require_operation_set
    def generate_yaml(self):
        from pyrasgo.api import Get
        return Get().dataset_yaml(self.id)

# --------
# Methods
# --------

    def transform(
            self,
            transform_name: str,
            arguments: Optional[Dict[str, Union[str, int, List, Dict, 'Dataset']]] = None,
            operation_name: Optional[str] = None,
            async_compute: Optional[bool] = True,
            render_only: Optional[bool] = False,
            **kwargs: Union[str, int, List, Dict, 'Dataset']
    ) -> Union['Dataset', None]:
        """
        Transform a new dataset with the given transform and arguments.
        Created operation is added to the dataset's canvas/operations set

        Args:
            transform_name: Name of transform to Apply
            arguments: Optional transform arguments sin not supplied by **kwargs
            operation_name: Name to set for the operation/transform
            render_only: Optional flag to simply render the operation that will result 
                from this transformation instead of using it to create a new Dataset.
            **kwargs:

        Returns:
             Returns an new dataset with the referenced transform
             added to this dataset's definition/operation set

             Optionally, can be used to try out some inputs to a transform by passing `render_only = True`
        """
        arguments = arguments if arguments else {}

        # Update the Transform arguments with any supplied kwargs
        arguments.update(kwargs)

        # Add required reference to self in transform
        arguments['source_table'] = self

        dataset_dependencies = set(self._dataset_dependencies)
        operation_dependencies = []
        parent_operations = []

        for k, v in arguments.items():
            if isinstance(v, self.__class__):
                if v._api_dataset:
                    v._assert_can_transform()
                    dataset_dependencies.add(v._api_dataset.id)
                else:
                    # if parent (could be self) is another transformed dataset, grab it's operations too
                    parent_operations.extend(v._operations)
                arguments[k] = v.fqtn
                operation_dependencies.append(v.fqtn)

        # Init table name GUID for outputted dataset
        new_guid = naming.random_operation_table_name()
        transform = self._get_transform_by_name(transform_name)

        operation_create = schemas.OperationCreate(
            operation_name=operation_name if operation_name else transform.name,
            operation_args=arguments,
            transform_id=transform.id,
            table_name=new_guid,
            table_dependency_names=operation_dependencies
        )

        if render_only:
            # Short circuit, and get the SQL we'll render for this operation.,
            from pyrasgo.api import Create
            print(Create()._operation_render(operation_create))
            return None

        operations = parent_operations + [operation_create]

        return self.__class__(
            operations=operations,
            dataset_dependencies=list(dataset_dependencies),
            guid=new_guid,
            async_compute=async_compute
        )

    @require_operation_set
    def to_df(
            self,
            filters: Optional[List[str]] = None,
            order_by: Optional[List[str]] = None,
            columns: Optional[List[str]] = None,
            limit: Optional[int] = None,
            snapshot_index: Optional[int] = None,
    ) -> pd.DataFrame:
        """
        Reads and returns this dataset into a pandas dataframe

        You can supply SQL WHERE clause filters, order the dataset by columns, only
        return selected columns, and add a return limit as well

        Example:
            ```
            ds = rasgo.get.dataset(dataset_id=74)
            ds.to_df(
                filters=['SALESTERRITORYKEY = 1', 'TOTALPRODUCTCOST BETWEEN 1000 AND 2000'],
                order_by=['TOTALPRODUCTCOST'],
                columns=['PRODUCTKEY', 'TOTALPRODUCTCOST', 'SALESTERRITORYKEY'],
                limit=50
            )
            ```

        Args:
            filters: List of SQL WHERE filters strings to filter on returned df
            order_by: List of columns to order by in returned dataset
            columns: List of columns to return in the df
            limit: Only return this many rows in the df
            snapshot_index: the index of a snapshot from Dataset.snapshots to read
        """
        from pyrasgo.api import Read
        return Read().dataset(
            dataset=self,
            filters=filters,
            order_by=order_by,
            columns=columns,
            limit=limit,
            snapshot_index=snapshot_index-1 if snapshot_index else None
        )

    def preview(
            self,
            filters: Optional[List[str]] = None,
            order_by: Optional[List[str]] = None,
            columns: Optional[List[str]] = None,
    ) -> pd.DataFrame:
        """
        Preview the first 10 rows of this dataset, returned as pandas dataframe

        You can supply SQL WHERE clause filters, order the dataset by columns, and
        only return selected columns

        Example:
            ```
            ds = rasgo.get.dataset(dataset_id=74)
            ds.preview(
                filters=['SALESTERRITORYKEY = 1', 'TOTALPRODUCTCOST BETWEEN 1000 AND 2000'],
                order_by=['TOTALPRODUCTCOST'],
                columns=['PRODUCTKEY', 'TOTALPRODUCTCOST', 'SALESTERRITORYKEY']
            )
            ```

        Args:
            filters: List of SQL WHERE filters strings to filter on returned df
            order_by: List of columns to order by in returned dataset
            columns: List of columns to return in the df
        """
        return self.to_df(filters, order_by, columns, limit=10)

# -----------------------------------------------------
# Methods requiring dataset to be registered with Rasgo
# -----------------------------------------------------

    def profile(self) -> any:
        """
        Get the URL for this Dataset in the RasgoUI
        """
        if not self.id:
            raise AttributeError("This dataset has not been created")
        
        from pyrasgo.api.session import Environment
        return f"{Environment.from_environment().app_path}/datasets/{self.id}"

    def run_stats(self, only_if_data_changed: Optional[bool] = True) -> None:
        """
        Trigger new stats for this datset

        Args:
            only_if_data_changed: Use False to tell Rasgo to run stats even if the data has not changed
        """
        if not self._api_dataset or not self._api_dataset.dw_table_id:
            raise APIError("Dataset must be created and have an output selected in order to generate stats")
        from pyrasgo.api.create import Create
        create = Create()
        
        table_id = self._api_dataset.dw_table_id

        create._dataset_correlation_stats(table_id=table_id, only_if_data_changed=only_if_data_changed)
        print(f"Request to generate stats for Dataset {self.id} started. Once generated, view stats at {self.profile()}")

# ---------------------------------
#  Private Helper Funcs for Class
# ---------------------------------

    def _get_transform_by_name(self, transform_name: str) -> schemas.Transform:
        """
        Get and return a transform obj by name

        Raise Error if no transform with that name found
        """
        for transform in self._available_transforms:
            if transform_name == transform.name:
                return transform
        raise ValueError(f"No Transform with name "
                         f"'{transform_name}' available to your organization")

    def _assert_can_transform(self) -> None:
        """
        Raise an API error if you can't transform this dataset
        """
        if self._api_dataset:
            if not self._api_dataset.dw_table_id or self.status != DS_PUBLISHED_STATUS:
                raise APIError(
                    f"Dataset({self._api_dataset.id}) has not been locked/published, "
                    f"and cannot be used as an arg for new dataset transformations"
                )

    def _require_operation_set(self) -> None:
        """
        This function used to ensure that an operation set exists for a given dataset before 
        attempting to do any operations that require the operations to exist in Rasgo (for example,
        previewing tables before they actually exist)

        If the operation set does not exist, AND the set of operations to be created does,
        create temp operations so the tables exists.

        NOTE: If this is a 'offline'/transformed dataset, the endpoint hit
        doesn't do everything to create a full working OP set for the UI for
        speed performance reason, just a temp one; Just enough so can preview
        datasets/source code. Call  `self._get_or_create_op_set()` or
        `rasgo.save.dataset()` to create the full working op set for UI and more
        """
        # If the dataset is from the API and it has a
        # op set, get and cache it if not done so yet
        if self._api_dataset:
            self._cache_op_set_from_api()

        # If this is a transformed/'offline' dataset create temp op set
        # and set source code preview
        if not self._source_code_preview and self._operations:
            from pyrasgo.api.create import Create
            if self._async_compute:
                task = Create()._operation_set_preview_async(
                    operations=self._operations,
                    dataset_dependency_ids=self._dataset_dependencies
                )
                self._source_code_preview = polling.poll_operation_set_offline_async_status(task)
            else:
                self._source_code_preview = Create()._operation_set_preview(
                    operations=self._operations,
                    dataset_dependency_ids=self._dataset_dependencies
                )

    def _cache_op_set_from_api(self) -> None:
        """
        If this dataset is from the API, and it's op set is not cached
        yet do so
        """
        if self._api_dataset:
            if not self._api_operation_set and self._api_dataset.dw_operation_set_id:
                from pyrasgo.api import Get
                self._api_operation_set = Get()._operation_set(
                    self._api_dataset.dw_operation_set_id
                )

    def _get_or_create_op_set(self) -> schemas.OperationSet:
        """
        Get or create plus return the operation set for this dataset
        """
        # If the dataset is from the API and it has a
        # op set, get and cache it if not done so yet
        if self._api_dataset:
            self._cache_op_set_from_api()

        # If the dataset doesn't have a op set in the API, or
        # it is a 'offline' transformed dataset create the op
        # set if not done so on this dataset yet
        if not self._api_operation_set:
            from pyrasgo.api.create import Create
            if self._async_compute:
                from pyrasgo.api.get import Get
                get = Get()

                # Submit the task request
                task_request = Create()._operation_set_async(
                    operations=self._operations,
                    dataset_dependency_ids=self._dataset_dependencies
                )
                operation_set_id = polling.poll_operation_set_async_status(task_request=task_request, verbose=self._verbose)
                self._api_operation_set = get._operation_set(operation_set_id)
            else:
                self._api_operation_set = Create()._operation_set(
                    operations=self._operations,
                    dataset_dependency_ids=self._dataset_dependencies
                )

        return self._api_operation_set

    def _create_udt_function(self, transform: schemas.Transform) -> Callable:
        """
        Creates and returns a new udt function to dynamically attached to the Dataset obj on init

        New funcs docstring, name, and signature (params shown when inspecting/doing . tab on func)
        as well to improve notebook experience of using UDTs for users

        Args:
            ds_transform_func: Function pointer of Dataset.transform()
            transform: Transform to read metadata and create new udt function for
        """

        # Create new function with 'transform_name` param set to this transform's name
        def f(*arg, **kwargs) -> 'Dataset':
            return self.transform(transform_name=transform.name, *arg, **kwargs)

        # Update func meta data for better inspection in notebook
        f.__name__ = transform.name
        f.__signature__ = _gen_udt_func_signature(f, transform)
        f.__doc__ = _gen_udt_func_docstring(transform)
        return f


# -----------------------------------------
#  Internal/private methods for this file
# -----------------------------------------


def _get_transforms() -> List[schemas.Transform]:
    """
    Get and set available transforms from the API to be used
    directly as functions of Dataset if not retrieved yet
    """
    # Get available transforms from the API to be used directly as functions of Dataset
    from pyrasgo.api import Get
    try:
        return Get().transforms()
    except:
        print('Unable to fetch available transforms from Rasgo.  '
              'Will not be able to transform this Dataset')
        return []


def _gen_udt_func_signature(udt_func: Callable, transform: schemas.Transform) -> inspect.Signature:
    """
    Creates and returns a UDT param signature.

    This is shown documentation for the parameters when hitting shift tab in a notebook
    """
    # Get current signature of function
    sig = inspect.signature(udt_func)

    # Create Signature Params for UDT Args
    udt_params = []
    for t_arg in transform.arguments:
        p = Parameter(name=t_arg.name, kind=Parameter.KEYWORD_ONLY)
        udt_params.append(p)

    # Add `operation_name` param as last in signature with type annotation
    op_name_param = Parameter(
        name='operation_name',
        kind=Parameter.KEYWORD_ONLY,
        annotation=Optional[str],
        default=None
    )
    udt_params.append(op_name_param)

    # Return new signature
    return sig.replace(parameters=udt_params)


def _gen_udt_func_docstring(transform: schemas.Transform) -> str:
    """
    Generate and return a docstring for a transform func
    with transform description, args, and return specified.
    """
    # Have start of docstring be transform description
    docstring = f"\n{transform.description}"

    # Add transform args to func docstring
    docstring = f"{docstring}\n  Args:"
    for t_arg in transform.arguments:
        docstring = f"{docstring}\n    {t_arg.name}: {t_arg.description}"
    docstring = f"{docstring}\n    operation_name: Name to set for the operation"

    # Add return to docstring
    docstring = f"{docstring}\n\n  Returns:\n    Returns an new dataset with the referenced " \
                f"{transform.name!r} added to this dataset's definition"
    return docstring



from pydantic import BaseModel, Field
from typing import List, Optional
from datetime import datetime

from pyrasgo.schemas.attributes import Attribute
from pyrasgo.schemas.dw import SimpleTable
from pyrasgo.schemas.granularity import Granularity

class DataSourceBase(BaseModel):
    id: int


class DataSourceColumn(BaseModel):
    id: int
    name: str
    dataType: str


class DimensionColumn(BaseModel):
    id: int
    columnName: str
    displayName: Optional[str]
    dataType: Optional[str]
    description: Optional[str]
    granularity: Optional[Granularity]
    dataSourceColumnId: Optional[int]


class FeatureColumn(BaseModel):
    id: int
    columnName: str
    displayName: Optional[str]
    dataType: Optional[str]
    status: Optional[str]
    description: Optional[str]
    tags: Optional[List[str]]
    attributes: Optional[List[Attribute]]
    gitRepo: Optional[str]
    dataSourceColumnId: Optional[int]
    collectionCount: Optional[int]


class DataSourceParent(DataSourceBase):
    id: int
    name: str
    table: Optional[str]
    tableDatabase: Optional[str]
    tableSchema: Optional[str]
    sourceType: Optional[str]
    domain: Optional[str]
    authorId: Optional[int]
    dataTable: Optional[SimpleTable]


class DataSource(DataSourceBase):
    id: int
    name: str
    table: Optional[str]
    tableDatabase: Optional[str]
    tableSchema: Optional[str]
    sourceCode: Optional[str]
    sourceType: Optional[str]
    domain: Optional[str]
    authorId: Optional[int]
    organizationId: int
    parentId: Optional[int]
    parentSource: Optional[DataSourceParent]
    columns: Optional[List[DataSourceColumn]]
    dimensions: Optional[List[DimensionColumn]]
    features: Optional[List[FeatureColumn]]
    dataTable: Optional[SimpleTable]


class DataSourceCreate(BaseModel):
    name: str
    table: Optional[str]
    tableDatabase: Optional[str]
    tableSchema: Optional[str]
    sourceCode: Optional[str]
    sourceType: Optional[str]
    domain: Optional[str]
    tableStatus: Optional[str]
    parentId: Optional[int]


class DataSourceUpdate(BaseModel):
    name: Optional[str]
    table: Optional[str]
    tableDatabase: Optional[str]
    tableSchema: Optional[str]
    sourceCode: Optional[str]
    sourceType: Optional[str]
    domain: Optional[str]
    tableStatus: Optional[str]
    parentId: Optional[int]


class DataSourceColumnPut(BaseModel):
    name: str
    dataType: str


class DimensionColumnPut(BaseModel):
    dataSourceColumnId: Optional[int]
    columnName: str
    displayName: Optional[str]
    description: Optional[str]
    granularityName: str
    dimensionType: Optional[str]


class FeatureColumnPut(BaseModel):
    dataSourceColumnId: Optional[int]
    columnName: str
    displayName: Optional[str]
    description: Optional[str]
    status: Optional[str]
    tags: Optional[List[str]]
    attributes: Optional[List[Attribute]]


class DataSourcePut(BaseModel):
    name: str
    table: str
    tableDatabase: Optional[str]
    tableSchema: Optional[str]
    sourceType: str
    domain: Optional[str]
    parentId: Optional[int]
    sourceCode: Optional[str]
    columns: List[DataSourceColumnPut]
    features: Optional[List[FeatureColumnPut]]
    dimensions: Optional[List[DimensionColumnPut]]
    organizationId: int


from .attributes import (Attribute, CollectionAttributeBulkCreate,
                         CollectionAttributes, CollectionAttributesLog,
                         FeatureAttributeBulkCreate, FeatureAttributes,
                         FeatureAttributesLog)
from .data_source import (DataSource, DataSourceColumn, DataSourceColumnPut,
                          DataSourceCreate, DataSourcePut, DataSourceUpdate,
                          DimensionColumnPut, FeatureColumnPut)
from .dataframe import Dataframe, DataframeCreate, DataframeUpdate
from .dataset import Dataset, DatasetBulk, DatasetCreate, DatasetUpdate
from .dataset_column import DatasetColumn, DatasetColumnCreate
from .dw_operation import Operation, OperationCreate
from .dw_operation_set import OperationSet, OperationSetCreate, OperationSetAsyncEvent, OperationSetAsyncTask, \
    OperationSetOfflineAsyncEvent, OperationSetOfflineAsyncTask
from .dw_table import DataColumn, DataTable, DataTableWithColumns
from .enums import DataType, Granularity, ModelType
from .feature import (ColumnProfiles, FeatureCreate, FeatureImportanceStats,
                      FeatureStats, FeatureUpdate)
from .organization import Organization
from .stats import GenerateStat
from .transform import (Transform, TransformArgumentCreate, TransformCreate,
                        TransformExecute, TransformUpdate)
from .user import User
from .yml import FeaturesYML



from enum import Enum
from pydantic import BaseModel
from typing import Optional

class ShareStatus(Enum):
    PRIVATE = 'private'
    ORGANIZATION = 'organization'
    PUBLIC = 'public'

class Dataframe(BaseModel):
    id: int
    name: Optional[str]
    uniqueId: str
    authorId: int
    authorName: Optional[str]
    organizationId: int
    sharedStatus: str
    columnHash: Optional[str]
    updatedDate: Optional[str]


class DataframeCreate(BaseModel):    
    name: Optional[str] = None
    uniqueId: str
    sharedStatus: str = 'organization'
    columnHash: Optional[str] = None
    updatedDate: Optional[str]


class DataframeUpdate(BaseModel):
    name: Optional[str]
    uniqueId: Optional[str]
    sharedStatus: Optional[str]
    columnHash: Optional[str]
    updatedDate: Optional[str]


from datetime import date, datetime
from pydantic import BaseModel, Field
from typing import Any, Dict, List, Optional, TypeVar, Union

from pyrasgo.schemas.attributes import Attribute
from pyrasgo.schemas.data_source import DataSourceParent, DimensionColumn
from pyrasgo.schemas.granularity import Granularity

UnknownValue = TypeVar('UnknownValue', float, int, str, date, datetime)

class FeatureBase(BaseModel):
    id: int


class FeatureCreate(BaseModel):
    name: str
    code: str
    description: Optional[str]
    dataSourceColumnId: Optional[int]
    dataSourceId: int
    orchestrationStatus: Optional[str]
    tags: Optional[List[str]]
    gitRepo: Optional[str]

class FeatureUpdate(BaseModel):
    id: int
    name: Optional[str]
    code: Optional[str]
    description: Optional[str]
    gitRepo: Optional[str]
    orchestrationStatus: Optional[str]

class Feature(FeatureBase):
    id: int
    displayName: Optional[str]
    columnName: str
    dataType: Optional[str]
    description: Optional[str]
    dataSourceColumnId: Optional[int]
    dataSource: Optional[DataSourceParent]
    dimensions: Optional[List[DimensionColumn]]
    granularities: Optional[List[Granularity]]
    orchestrationStatus: Optional[str] = Field(alias="status")
    tags: Optional[List[str]]
    attributes: Optional[List[Attribute]]
    gitRepo: Optional[str]
    organizationId: Optional[int]
    class Config:
        allow_population_by_field_name = True

class Bucket(BaseModel):
    bucketCeiling: Optional[UnknownValue]
    bucketCeilingLabel: Optional[str]
    bucketFloor: Optional[UnknownValue]
    bucketFloorLabel: Optional[str]
    beginVal: Optional[UnknownValue]
    endVal: Optional[UnknownValue]
    height: Optional[UnknownValue]
    histBucket: Optional[Union[int, str]]
    outlierCt: Optional[UnknownValue]


class Value(BaseModel):
    freq: Optional[float]
    isOutlier: Optional[int]
    recCt: Optional[int]
    val: Optional[UnknownValue]


class FeatureStats(BaseModel):
    recCt: Optional[int]
    distinctCt: Optional[int]
    nullRecCt: Optional[int]
    zeroValRecCt: Optional[int]
    meanVal: Optional[str]
    medianVal: Optional[str]
    maxVal: Optional[str]
    minVal: Optional[str]
    sumVal: Optional[str]
    stdDevVal: Optional[str]
    varianceVal: Optional[str]
    rangeVal: Optional[str]
    skewVal: Optional[str]
    kurtosisVal: Optional[str]
    q1Val: Optional[str]
    q3Val: Optional[str]
    IQRVal: Optional[str]
    pct5Val: Optional[str]
    pct95Val: Optional[str]
    outlierCt: Optional[str]
    lowOutlier: Optional[str]
    highOutlier: Optional[str]


class ColumnStats(BaseModel):
    columnName: Optional[str]
    dataType: Optional[str]
    featureGranularity: Optional[str]
    featureId: Optional[str]
    featureName: Optional[str]
    featureStats: Optional[FeatureStats]
    histogram: Optional[List[Bucket]]
    commonValues: Optional[List[Value]]


class ColumnProfiles(BaseModel):
    columnProfiles: Optional[List[ColumnStats]]
    timestamp: str


class modelPerformance(BaseModel):
    R2: Optional[float]
    RMSE: Optional[float]
    AUC: Optional[float]
    Logloss: Optional[float]
    Precision: Optional[float]
    Recall: Optional[float]


class FeatureImportanceStats(BaseModel):
    targetFeature: str
    featureShapleyDistributions: dict
    featureImportance: Dict[str, UnknownValue]
    modelPerformance: modelPerformance
    timestamp: str
    modelId: Optional[int]



"""
API Contracts for DW Operation Sets
"""
from datetime import datetime

from pydantic import BaseModel, Field
from typing import List, Optional, Any

from pyrasgo.schemas import dw_operation as op_contracts
from pyrasgo.schemas import dataset as datasets_contracts


class OperationSetCreate(BaseModel):
    """
    Contract to create an operation set
    """
    operation_type: str = Field(default="udt", alias="operationType")
    operations: Optional[List[op_contracts.OperationCreate]]
    dataset_dependency_ids: Optional[List[int]] = Field(alias="datasetDependencyIds") # Dataset Ids

    class Config:
        allow_population_by_field_name = True


class OperationSetUpdate(BaseModel):
    """
    Contract to update an operation set
    """
    operation_type: str = Field(default="udt", alias="operationType")
    operations: Optional[List[op_contracts.OperationUpdate]]
    dataset_dependency_ids: Optional[List[int]] = Field(alias="datasetDependencyIds") # Dataset Ids

    class Config:
        allow_population_by_field_name = True


class OperationSet(BaseModel):
    """
    Contract to return from get endpoints
    """
    id: int
    operation_type: str = Field(default="udt", alias="operationType")
    operations: Optional[List[op_contracts.Operation]]
    dataset_dependencies: Optional[List[datasets_contracts.Dataset]] = Field(alias="datasetDependencies")
    organization_id: int = Field(alias="organizationId")

    class Config:
        allow_population_by_field_name = True


class OperationSetAsyncEvent(BaseModel):
    """
    Represents a single event that happened while executing an async OperationSetCreate task
    """
    id: int
    create_timestamp: datetime = Field(alias='createTimestamp')
    event_type: str = Field(alias='eventType')
    message: Optional[Any]


class OperationSetAsyncTask(BaseModel):
    """
    Response model for when a caller asks to create a task or polls for that task's status
    """
    id: int = Field()
    request: OperationSetCreate = Field()
    create_author: Optional[int] = Field(alias='createAuthor')
    organization_id: int = Field(alias='organizationId')
    events: List[OperationSetAsyncEvent]


class OperationSetOfflineAsyncEvent(BaseModel):
    """
    Represents a single event that happened while executing an async OperationSetCreate offline task
    """
    id: int
    create_timestamp: datetime = Field(alias='createTimestamp')
    event_type: str = Field(alias='eventType')


class OperationSetOfflineAsyncTask(BaseModel):
    """
    Response model for when a caller asks to create a task or polls for that task's status
    """
    id: int = Field()
    request: Any = Field()
    create_author: Optional[int] = Field(alias='createAuthor')
    organization_id: int = Field(alias='organizationId')
    events: List[OperationSetAsyncEvent]
